/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Documentation - RestDocsDocumentation.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.test.doc;

import static ch.unige.solidify.test.executor.SolidifyExecutorTestDoc.makeFieldsDocumentation;
import static ch.unige.solidify.test.executor.SolidifyExecutorTestDoc.makeLinksDocumentation;
import static ch.unige.solidify.test.executor.SolidifyExecutorTestDoc.performExpectationTests;
import static ch.unige.solidify.test.executor.SolidifyExecutorTestDoc.performRequest;
import static ch.unige.solidify.test.helper.RestDocsTemplates.HREF_FIELD;
import static ch.unige.solidify.test.helper.RestDocsTemplates.LINKS_FIELD;
import static ch.unige.solidify.test.helper.RestDocsType.APPLICATION;
import static ch.unige.solidify.test.helper.RestDocsType.MODULE;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.fail;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.documentationConfiguration;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import org.json.JSONObject;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpServletResponse;
import org.springframework.restdocs.RestDocumentationContextProvider;
import org.springframework.restdocs.RestDocumentationExtension;
import org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.context.WebApplicationContext;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.controller.ModuleController;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.test.exception.SolidifyTestRuntimeException;
import ch.unige.solidify.test.helper.AbstractResourceDoc;
import ch.unige.solidify.test.helper.RestDocsTemplates;
import ch.unige.solidify.test.helper.RestDocsType;
import ch.unige.solidify.test.model.OpenApiExtraDocumentation;

/**
 * Generic Rest-Docs context configuration for Solidify applications
 */
@ExtendWith(RestDocumentationExtension.class)
public abstract class RestDocsDocumentation {

  private static final String DATA_FIELD = "_data";

  private static final Logger logger = LoggerFactory.getLogger(RestDocsDocumentation.class);
  private static final String RES_ID_FIELD = "resId";
  private static final String URL_SEPARATOR = "/";

  protected Gson gson;
  protected List<String> urlIgnoreList = new ArrayList<>();
  protected List<String> urlReadOnlyList = new ArrayList<>();

  @Value("${server.servlet.context-path}")
  private String contextPath;

  private final Map<String, String> ids = new HashMap<>();

  private final List<OpenApiExtraDocumentation> listOpenApiExtraDocumentation = new ArrayList<>();

  private final Logger log = LoggerFactory.getLogger(RestDocsDocumentation.class);

  private MockMvc mockMvc;

  @Autowired
  private List<ModuleController> moduleControllerList;

  private final Map<URI, AbstractResourceDoc> resources = new HashMap<>();

  @Autowired
  private WebApplicationContext webApplicationContext;

  @Autowired
  RestDocsTemplates restDocsTemplates;

  /**
   * The main REST calls of each application
   */
  @Test
  public void generateMainRESTDocs() {
    final ResultActions result = this.executeMockRequest(APPLICATION, null, RequestMethod.GET);
    performExpectationTests(result, this.restDocsTemplates.getApplicationTestTemplate(this.getModules()));
    final String uri = result.andReturn().getRequest().getRequestURI();
    makeLinksDocumentation(result, uri, false, this.restDocsTemplates.getApplicationLinkDescriptors(this.getModules()));
    makeFieldsDocumentation(result, uri, this.restDocsTemplates.getApplicationFieldsTemplate(), true);
  }

  /**
   * The main REST calls of each module
   */
  @Test
  public void generateModulesRESTDocs() throws IOException {
    for (final String module : this.getModules()) {
      if (this.isIgnoredUrl(this.getModuleUrl(module))) {
        this.log.warn("Ignored module {}", module);
        continue;
      }
      final ResultActions result = this.executeMockRequest(MODULE, module, RequestMethod.GET);
      final String uri = result.andReturn().getRequest().getRequestURI();
      final List<String> resourcesFromResult = this.fetchResources(result, RestDocsTemplates.RESOURCES_FIELD);
      final List<String> toolsFromResult = this.fetchResources(result, RestDocsTemplates.TOOLS_FIELD);
      if (!resourcesFromResult.isEmpty()) {
        makeLinksDocumentation(result, uri, false,
                this.restDocsTemplates.getModulesLinkDescriptors(module, resourcesFromResult, toolsFromResult));
        makeFieldsDocumentation(result, uri, this.restDocsTemplates.getModuleFieldsDescriptor(module), true);
      }
    }
  }

  /**
   * The main REST calls of each resource
   */
  @Test
  public void generateResourcesRESTDocs() throws IOException {
    for (final String module : this.getModules()) {
      if (this.isIgnoredUrl(this.getModuleUrl(module))) {
        this.log.warn("Ignored module {}", module);
        continue;
      }
      final ResultActions resultActions = this.executeMockRequest(MODULE, module, RequestMethod.GET);
      // Data resource generation
      this.fetchResources(resultActions, RestDocsTemplates.RESOURCES_FIELD).stream()
              .forEach(resource -> this.generateResourceRESTDocs(module, resource));
    }
    this.generateOpenApiExtraDocumentation();
  }

  @BeforeEach
  public void setUp(RestDocumentationContextProvider restDocumentation) {
    this.gson = new GsonBuilder().create();
    this.mockMvc = MockMvcBuilders.webAppContextSetup(this.webApplicationContext).defaultRequest(this.getApplicationRequestBuilder())
            .apply(documentationConfiguration(restDocumentation)).build();
    this.setUrlIgnoreList();
    this.log.info("Total ignored URLs {}", this.urlIgnoreList.size());
    this.setUrlReadOnlyList();
    this.log.info("Total Read-Only URLs {}", this.urlReadOnlyList.size());
    this.loadResources();
  }

  protected void addResources(AbstractResourceDoc r) {
    this.resources.put(URI.create(this.contextPath + r.getName()), r);
  }

  /**
   * @return the context path
   */
  protected String getContextPath() {
    return this.contextPath;
  }

  protected abstract void loadResources();

  /**
   * Used to bypass no testable url's.
   */
  protected abstract void setUrlIgnoreList();

  protected abstract void setUrlReadOnlyList();

  private void createDirectory(String outputFilePath) {
    final File directory = new File(this.getDirectoryPath(outputFilePath));
    if (!directory.exists()) {
      directory.mkdirs();
    }
  }

  /**
   * Build and execute a mock request
   */
  private ResultActions executeMockRequest(RestDocsType restDocsType, String path, RequestMethod method) {
    return this.executeMockRequest(restDocsType, path, method, null);
  }

  private void generateResourceRESTDocs(String module, String resource) {
    final URI uri = URI.create(URI.create(resource).getPath());
    final String resourcePath = this.getResourcePath(resource);
    ResultActions listResult;
    if (!this.isIgnoredUrl(resourcePath)) {
      // Test if doc exists
      if (this.resources.get(uri) != null) {
        final AbstractResourceDoc docTemplate = this.resources.get(uri);
        // Test if normal resource
        if (docTemplate.isNormalResource()) {

          this.documenteNormalResource(module, resource, docTemplate);
        } else {
          // Not standard resource
          this.documenteNotNormalResource(resource, docTemplate);
        }
        // Log entity for OpenApi
        this.registerExtraDocumentation(docTemplate.getOpenApiExtraDocumentation());
      } else {
        // Resource to document
        this.log.error("No documentation for {}", resourcePath);
        // List resource collection
        listResult = this.executeMockRequest(RestDocsType.RESOURCE_LIST, resource, RequestMethod.GET);
        makeFieldsDocumentation(listResult, resourcePath, this.restDocsTemplates.getResourceCollectionFieldDescriptors(),
                this.isReadOnlyUrl(resourcePath));
      }
    } else {
      this.log.warn("Ignored resource {}", resourcePath);
    }
  }

  private void documenteNotNormalResource(String resource, AbstractResourceDoc docTemplate) {
    String resourcePath = this.getResourcePath(resource);
    // List resource collection
    ResultActions resourceResult = this.executeMockRequest(RestDocsType.RESOURCE, resource, RequestMethod.GET, docTemplate);
    // Document resource fields
    makeFieldsDocumentation(resourceResult, resourcePath,
            this.restDocsTemplates.getResourceFieldDescriptors(resourcePath, docTemplate.getReadOnlyDocumentationFields(),
                    docTemplate.getCommonResourceFields()),
            this.isReadOnlyUrl(resourcePath));
  }

  private String getResourcePath(String resource) {
    final URI uri = URI.create(URI.create(resource).getPath());
    return uri.getPath();
  }

  private void documenteNormalResource(String module, String resource, AbstractResourceDoc docTemplate) {
    ResultActions resourceResult;
    String resourcePath = this.getResourcePath(resource);
    if (!this.isReadOnlyUrl(resourcePath)) {
      this.log.info("Document resource {}", resourcePath);
      // Create a resource
      resourceResult = this.executeMockRequest(RestDocsType.RESOURCE, resource, RequestMethod.POST, docTemplate);
    } else {
      this.log.info("Document read-only resource {}", resourcePath);
      // Find first resource (Read-Only)
      MockHttpServletResponse listResponse = this.executeMockRequest(RestDocsType.RESOURCE_LIST, resource, RequestMethod.GET, docTemplate)
              .andReturn().getResponse();
      try {
        final String resId = this.getFirstResId(listResponse.getContentAsString());
        resourceResult = this.executeMockRequest(RestDocsType.RESOURCE, resource + SolidifyConstants.URL_SEP + resId, RequestMethod.GET,
                docTemplate);
      } catch (UnsupportedEncodingException e) {
        throw new SolidifyRuntimeException(e.getMessage());
      }
    }
    // List resource collection
    ResultActions listResult = this.executeMockRequest(RestDocsType.RESOURCE_LIST, resource, RequestMethod.GET);
    // Document resource fields
    makeFieldsDocumentation(resourceResult, resourcePath,
            this.restDocsTemplates.getResourceFieldDescriptors(resourcePath, docTemplate.getReadOnlyDocumentationFields(),
                    docTemplate.getCommonResourceFields()),
            this.isReadOnlyUrl(resourcePath));
    // Document resource links
    makeLinksDocumentation(resourceResult, resourcePath, false,
            this.restDocsTemplates.getResourceDataLinkDescriptors(resourcePath, docTemplate.getResourceLinks()));
    // Document resource collection links
    makeLinksDocumentation(listResult, resourcePath, true,
            this.restDocsTemplates.getResourceLinkDescriptors(module, resourcePath, docTemplate.getListLinks()));

  }

  private ResultActions executeMockRequest(RestDocsType restDocsType, String path, RequestMethod method, AbstractResourceDoc resDoc) {
    MockHttpServletRequestBuilder mockHttpServletRequestBuilder = null;
    ResultActions result = null;
    int expectedSize = 0;
    this.log.info("Document {}", path);
    switch (restDocsType) {
      case RESOURCE_LIST:
        mockHttpServletRequestBuilder = this.getResourceRequestBuilder(path, RequestMethod.GET, null);
        expectedSize = this.restDocsTemplates.getCommonResourceListFields().size();
        break;
      case RESOURCE:
        if (resDoc == null) {
          throw new SolidifyRuntimeException("Missing resource documentation");
        }
        if (RequestMethod.POST.equals(method)) {
          resDoc.updateValues(this.ids);
          mockHttpServletRequestBuilder = this.getResourceRequestBuilder(path, RequestMethod.POST, resDoc.getJsonFields());
        } else {
          mockHttpServletRequestBuilder = this.getResourceRequestBuilder(path, RequestMethod.GET, null);
        }
        expectedSize += resDoc.getNumberOfExpectedReadOnlyFields() + resDoc.getCommonResourceFields().size();
        break;
      case MODULE:
        mockHttpServletRequestBuilder = this.getModuleRequestBuilder(path);
        expectedSize = this.restDocsTemplates.getCommonModuleFields().size();
        break;
      case APPLICATION:
        mockHttpServletRequestBuilder = this.getApplicationRequestBuilder();
        expectedSize = this.restDocsTemplates.getCommonApplicationFields().size();
        break;
    }
    try {
      result = performRequest(this.mockMvc, mockHttpServletRequestBuilder, method);
      final JsonNode jsonMockResponse = new ObjectMapper().readTree(result.andReturn().getResponse().getContentAsString());
      // Add resource id to map of ids
      if (restDocsType == RestDocsType.RESOURCE) {
        if (resDoc == null) {
          throw new SolidifyTestRuntimeException("Resource documentation must not be null");
        }
        if (resDoc.isNormalResource()) {
          this.ids.put(resDoc.getClass().getSimpleName(), jsonMockResponse.get(RES_ID_FIELD).textValue());
        }
      }
      assertEquals(expectedSize, jsonMockResponse.size());
    } catch (final IOException e) {
      fail("[" + path + "] Failure getting json response content. [" + result.andReturn().getResponse() + "]");
    }
    return result;
  }

  /**
   * Fetch resources from json response
   */
  private List<String> fetchResources(ResultActions resultActions, String resourceType) throws IOException {
    final JsonNode json = new ObjectMapper().readTree(resultActions.andReturn().getResponse().getContentAsString());
    if (json.path(LINKS_FIELD).get(resourceType) == null) {
      return new ArrayList<>();
    }
    List<String> resourceList = json.path(LINKS_FIELD).get(resourceType).findValuesAsText(HREF_FIELD);
    resourceList.sort(String::compareTo);
    return resourceList;
  }

  private void generateOpenApiExtraDocumentation() {
    final String jsonExtraDocumentation = this.gson.toJson(this.listOpenApiExtraDocumentation);
    try {
      this.saveJsonFile("target/restdocs-spec/openapi-extra-documentation.json", jsonExtraDocumentation);
    } catch (final IOException e) {
      this.log.error("IO exception", e);
      fail("IO exception: " + e.getMessage());
    }
  }

  /**
   * Return a request builder for applications
   */
  private MockHttpServletRequestBuilder getApplicationRequestBuilder() {
    return RestDocumentationRequestBuilders.get(this.contextPath).contextPath(this.contextPath);
  }

  private String getDirectoryPath(String outputFilePath) {
    if (outputFilePath.endsWith(URL_SEPARATOR)) {
      return outputFilePath;
    }
    final int lastIndexOfSeparator = outputFilePath.lastIndexOf(URL_SEPARATOR);
    return outputFilePath.substring(0, lastIndexOfSeparator);
  }

  private String getFirstResId(String contentAsString) {
    final ObjectMapper mapper = new ObjectMapper();
    mapper.registerModule(new com.fasterxml.jackson.datatype.jsr310.JavaTimeModule());
    mapper.configure(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS, false);
    mapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

    try {
      final JsonNode jsonData = mapper.readTree(contentAsString).get(DATA_FIELD);
      if (jsonData.isArray()) {
        for (final JsonNode json : jsonData) {
          if (json.get(RES_ID_FIELD) != null) {
            return json.get(RES_ID_FIELD).textValue();
          }
        }
      }

    } catch (final IOException e) {
      throw new SolidifyRuntimeException(e.getMessage());
    }
    throw new SolidifyRuntimeException("Cannot find first resId");
  }

  /**
   * Return a request builder for modules
   */
  private MockHttpServletRequestBuilder getModuleRequestBuilder(String module) {
    final String url = this.getModuleUrl(module);
    this.log.info("Building module level request [{}]", url);
    final MockHttpServletRequestBuilder mockHttpServletRequestBuilder = RestDocumentationRequestBuilders.get(url);
    mockHttpServletRequestBuilder.contextPath(this.contextPath);
    return mockHttpServletRequestBuilder;
  }

  /**
   * Get the modules list corresponding with the current application
   */
  private List<String> getModules() {
    return this.moduleControllerList.stream()
            .map(ModuleController::getRoot)
            .collect(Collectors.toList());
  }

  private String getModuleUrl(String module) {
    return this.contextPath + SolidifyConstants.URL_SEP + module;
  }

  /**
   * Return a request builder for resourcesTestObjectBuilder
   */
  private MockHttpServletRequestBuilder getResourceRequestBuilder(String resource, RequestMethod method, JSONObject content) {
    this.log.info("Building resource level request [{}], method=[{}]", resource, method);
    MockHttpServletRequestBuilder mockHttpServletRequestBuilder;
    if (method.equals(RequestMethod.POST) && content != null) {
      mockHttpServletRequestBuilder = RestDocumentationRequestBuilders
              .post(resource).content(content.toString()).contentType(MediaType.APPLICATION_JSON);
    } else {
      assertNull(content);
      mockHttpServletRequestBuilder = RestDocumentationRequestBuilders.get(resource);
    }
    return mockHttpServletRequestBuilder.contextPath(this.contextPath);
  }

  private boolean isIgnoredUrl(String url) {
    return this.isInList(this.urlIgnoreList, url);
  }

  private boolean isReadOnlyUrl(String url) {
    return this.isInList(this.urlReadOnlyList, url);
  }

  private boolean isInList(List<String> list, String url) {
    final Optional<String> optional = list.stream().filter(s -> s.equals(url)).findFirst();
    return optional.isPresent();
  }

  private void registerExtraDocumentation(OpenApiExtraDocumentation openApiExtraDocumentation) {
    this.listOpenApiExtraDocumentation.add(openApiExtraDocumentation);
  }

  private void saveJsonFile(String outputFilePath, String json) throws IOException {
    this.createDirectory(outputFilePath);
    final Path jsonFile = Paths.get(outputFilePath);
    // Use try-with-resource to get auto-closeable writer instance
    try (BufferedWriter writer = Files.newBufferedWriter(jsonFile)) {
      writer.write(json);
    }
    logger.info("Successfully Copied JSON Object into : {}", outputFilePath);
  }

}
