/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Email - EmailService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.service;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import javax.mail.MessagingException;
import javax.mail.internet.MimeMessage;

import org.jsoup.Jsoup;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.mail.MailException;
import org.springframework.mail.SimpleMailMessage;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.mail.javamail.MimeMessageHelper;
import org.springframework.stereotype.Service;
import org.thymeleaf.context.Context;
import org.thymeleaf.spring5.SpringTemplateEngine;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.config.SolidifyProperties;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.util.StringTool;

@Service
public class EmailService {
  private static final Logger log = LoggerFactory.getLogger(EmailService.class);

  private JavaMailSender emailSender;
  private SpringTemplateEngine templateEngine;
  private String senderAddress;
  private List<String> ccList = new ArrayList<>();
  private List<String> bccList = new ArrayList<>();

  public EmailService(SolidifyProperties config, JavaMailSender emailSender, SpringTemplateEngine templateEngine) {
    // Inject beans
    this.emailSender = emailSender;
    this.templateEngine = templateEngine;
    // Sender email
    if (StringTool.isNullOrEmpty(config.getEmail().getSenderAddress())) {
      throw new SolidifyRuntimeException("Missing sender address ('solidify.email.sender-address' parameter)");
    }
    this.senderAddress = config.getEmail().getSenderAddress();
    // List (CC & BCC) to copy
    if (config.getEmail().getCcList().length > 0) {
      this.ccList = Arrays.asList(config.getEmail().getCcList());
    }
    if (config.getEmail().getBccList().length > 0) {
      this.bccList = Arrays.asList(config.getEmail().getBccList());
    }
  }

  // ********************
  // ** Public methods **
  // ********************

  // HTML emails
  public void sendEmailWithTemplate(String template, List<String> toList, String subject, Map<String, Object> parameters)
          throws MessagingException {
    this.sendEmailWithTemplate(template, toList, null, null, subject, parameters);
  }

  public void sendEmailWithTemplate(String template, List<String> toList, List<String> ccList, List<String> bccList, String subject,
          Map<String, Object> parameters)
          throws MessagingException {
    final Context thymeleafContext = new Context();
    thymeleafContext.setVariables(parameters);
    String htmlBody = this.templateEngine.process(template + SolidifyConstants.HTML_EXT, thymeleafContext);
    this.sendHtmlEmail(this.senderAddress, toList, ccList, bccList, subject, htmlBody);
  }

  // Text email
  public void sendEmail(List<String> toList, String subject, String body) {
    this.sendEmail(toList, null, null, subject, body);
  }

  public void sendEmail(List<String> toList, List<String> ccList, List<String> bccList, String subject, String body) {
    this.sendTextEmail(this.senderAddress, toList, ccList, bccList, subject, body);
  }

  // *********************
  // ** Private methods **
  // *********************
  private void sendTextEmail(String from, List<String> toList, List<String> ccList, List<String> bccList, String subject, String body) {
    try {
      SimpleMailMessage email = new SimpleMailMessage();
      // From
      email.setFrom(from);
      // To
      email.setTo(this.toArray(toList));
      // CC
      ccList = this.completeList(ccList, this.ccList);
      if (!ccList.isEmpty()) {
        email.setCc(this.toArray(ccList));
      }
      // BCC
      bccList = this.completeList(bccList, this.bccList);
      if (!bccList.isEmpty()) {
        email.setBcc(this.toArray(bccList));
      }
      // Subject
      email.setSubject(subject);
      // Body
      email.setText(body);
      // Sent email
      this.emailSender.send(email);
      log.info("Email sent: {}", email);
    } catch (MailException e) {
      log.error("Email sending error: {}", e.getMessage(), e);
    }
  }

  private void sendHtmlEmail(String from, List<String> toList, List<String> ccList, List<String> bccList, String subject, String htmlBody)
          throws MessagingException {
    MimeMessage email = this.emailSender.createMimeMessage();
    MimeMessageHelper emailHelper = new MimeMessageHelper(email, true, "UTF-8");
    // From
    emailHelper.setFrom(from);
    // To
    emailHelper.setTo(this.toArray(toList));
    // CC
    ccList = this.completeList(ccList, this.ccList);
    if (!ccList.isEmpty()) {
      emailHelper.setCc(this.toArray(ccList));
    }
    // BCC
    bccList = this.completeList(bccList, this.bccList);
    if (!bccList.isEmpty()) {
      emailHelper.setBcc(this.toArray(bccList));
    }
    // Subject
    emailHelper.setSubject(subject);
    // Convert HTML in text
    String plainBody = this.toPlainText(htmlBody);
    // Add body in text & HTML
    emailHelper.setText(plainBody, htmlBody);
    // Send email
    this.emailSender.send(email);
  }

  private String toPlainText(String htmlText) {
    String plainText = Jsoup.parse(htmlText).wholeText();
    StringBuilder sb = new StringBuilder();
    for (String line : plainText.split("\n")) {
      if (!StringTool.isNullOrEmpty(line.strip())) {
        sb.append(line.strip());
        sb.append("\n");
      }
    }
    return sb.toString();
  }

  private String[] toArray(List<String> list) {
    if (list == null) {
      list = new ArrayList<>();
    }
    return list.toArray(new String[list.size()]);
  }

  private List<String> completeList(List<String> list, List<String> copyList) {
    if (list == null) {
      list = new ArrayList<>();
    }
    if (!copyList.isEmpty()) {
      list.addAll(copyList);
    }
    return list;
  }
}
